
var exec = require('child_process').exec;
var trace;
var flexsettings;
var session_data = {};
var path = require('path');
var fs   = require('fs');

module.exports = function(settings, trace_obj)
{
  flexsettings = settings;
  trace = trace_obj;
  return GetInst(  );
};


function GetInst(  ) {
    var _inst = {
        Exec( msg, callback) {
            if ( !msg.tool ) return callback(msg.verb + ": Invalid request. Required parameter 'tool' is missing");
            var tool_path = flexsettings.GetSetting("tool_path_"+msg.tool, '');
            if ( !tool_path ) return callback(msg.verb + ": A tool " + msg.tool + " is not supported");
            tool_path = PathReplacements(tool_path);
            if ( tool_path.indexOf("..") >= 0) return callback(msg.verb + ": A tool path can't be relative! " + tool_path);
            if ( fs.existsSync(tool_path) == false ) return callback("A tool not found: " + tool_path);

            var tool_params = (msg.params_id)? flexsettings.GetSetting("tool_params_"+ msg.params_id+ "_" +  msg.tool, ''): "";
            tool_params = ParamsReplacements( tool_params, msg );
            tool_params = PathReplacements(tool_params);

            var cwd = msg.cwd || "{data}/" + msg.tool;
            cwd = PathReplacements(cwd);
            if ( fs.existsSync(cwd) == false ) fs.mkdirSync(cwd);

            if ( msg.file_data && msg.file_name) {
                var filepath = PathForRunningOS(path.join(flexsettings.GetDataPath(), "uploads", msg.file_name ));
               var data_to_save = (msg.is_text)?msg.file_data : Buffer.from(msg.file_data);
                fs.writeFileSync( filepath, data_to_save);
            }
            var session_id = ExecChildProcess( tool_path, cwd, tool_params );
            //clone request excluding file_data and adding status and session_id fields to the response
            const {file_data, ...resp } = {"status" : "ok", "session_id" : session_id, ...msg};
            callback( null, resp );
        },

        GetCurrentStatus( msg, callback ){
            if ( session_data[msg.session_id] == null ) callback( "No data found by session id " + msg.session_id);
            else callback( null, { ...msg, status : "ok",  ...session_data[msg.session_id]} );
        }
   }
   return _inst;
}


function ExecChildProcess ( executor, cwd, params_str, callback ){
    var session_id = new Date().getTime();
    var  options = { cwd: cwd };
    session_data[session_id] = { stdout : "", stderr : "", finished : false };
    trace.Info("Run " + '"'+executor + '"' + ' ' + params_str);
    var child = exec( '"'+executor + '"' + ' ' + params_str, options);

    child.stdout.on('data', function(data) {
        session_data[session_id].stdout += data;
    });
    child.stderr.on('data', function(data) {
        trace.Error('Error when execute: ' + executor + ' '+params_str + ' stderr: ' + data);
        session_data[session_id].stderr = data;
    });
    child.on('close', function(code) {
        trace.Info("Process " + executor + " finished with code " + code);
        session_data[session_id].finished = true;
        session_data[session_id].err_code = code;
        if ( callback != null ) callback(code, session_data[session_id]);
    });
    return session_id;
}

function PathForRunningOS( filepath, specific_sep ) {
    var sep = path.sep;
    if (typeof specific_sep !== 'undefined') {
        sep = specific_sep;
    }
    // console.log("  PathForRunningOS --checking filepath: " + filepath );
    if (typeof filepath === "string") {
        if (( filepath.indexOf('\\') > -1) || ( filepath.indexOf('/') > -1)) {
            if ( sep == '/') {      // adjust for Linux
                filepath = filepath.split('\\').join('/');
            } else {                // adjust for Windows
                filepath = filepath.split('/').join('\\');
            }
        }
    }
    return filepath;
}

function PathReplacements(str)
{
    str = str.replace('{data}',flexsettings.GetDataPath());
    str = str.replace('{app-lib-bin}', path.resolve(__dirname, "../app-lib/bin"));
    str = str.replace('{uploads}', path.join(flexsettings.GetDataPath(), "uploads"));
    return PathForRunningOS(str);
}

function GetVars(template)
{
    var i = 0;
    var data = [];

    do {
        if (template[i] == '%') {
            for (var j=i+1; j<template.length; j++) {
                if (template[j] == '%') {
                data[data.length] = template.slice(i+1, j);
                i = j+1;
                break;
                }
            }
        }
    } while (++i < template.length);
  return data;
}

// replaces any variables %xxx% in str by value from req.xxx
function ParamsReplacements(str, req)
{
    var retval = str;
    var vars = GetVars(str);
    vars.forEach((v)=>{
        if ( v in req ) retval = retval.replace('%'+v+'%', req[v]);
        else retval = retval.replace('%'+v+'%', "");
    });
    return retval;
}